@@ -4,12 +4,27 @@ Types for high level PKCS#11 wrapper.
 This module provides stubs that are overrideen in pkcs11._pkcs11.
 """
 
+import struct
 from threading import RLock
 from binascii import hexlify
 
-from .constants import *
-from .mechanisms import *
-from .exceptions import *
+from .constants import (
+    Attribute,
+    MechanismFlag,
+    ObjectClass,
+    SlotFlag,
+    TokenFlag,
+    UserType,
+)
+from .mechanisms import Mechanism
+from .exceptions import (
+    ArgumentsBad,
+    AttributeTypeInvalid,
+    NoSuchKey,
+    MultipleObjectsReturned,
+    SignatureInvalid,
+    SignatureLenRange,
+)
 
 
 def _CK_UTF8CHAR_to_str(data):
@@ -32,6 +47,40 @@ def _CK_MECHANISM_TYPE_to_enum(mechanism):
         return mechanism
 
 
+class MechanismInfo:
+    """
+    Information about a mechanism.
+
+    See :meth:`pkcs11.Slot.get_mechanism_info`.
+    """
+
+    def __init__(self,
+                 slot,
+                 mechanism,
+                 ulMinKeySize=None,
+                 ulMaxKeySize=None,
+                 flags=None,
+                 **kwargs):
+
+        self.slot = slot
+        """:class:`pkcs11.Slot` this information is for."""
+        self.mechanism = mechanism
+        """:class:`pkcs11.mechanisms.Mechanism` this information is for."""
+        self.min_key_length = ulMinKeySize
+        """Minimum key length in bits (:class:`int`)."""
+        self.max_key_length = ulMaxKeySize
+        """Maximum key length in bits (:class:`int`)."""
+        self.flags = MechanismFlag(flags)
+        """Mechanism capabilities (:class:`pkcs11.constants.MechanismFlag`)."""
+
+    def __str__(self):
+        return '\n'.join((
+            "Supported key lengths: [%s, %s]" % (self.min_key_length,
+                                                 self.max_key_length),
+            "Flags: %s" % self.flags,
+        ))
+
+
 class Slot:
     """
     A PKCS#11 device slot.
@@ -80,6 +129,15 @@ class Slot:
         """
         raise NotImplementedError()
 
+    def get_mechanism_info(self, mechanism):
+        """
+        Returns information about the mechanism.
+
+        :param Mechanism mechanism: mechanism to learn about
+        :rtype: MechanismInfo
+        """
+        raise NotImplementedError()
+
     def __eq__(self, other):
         return self.slot_id == other.slot_id
 
@@ -303,10 +361,26 @@ class Session:
         """
         raise NotImplementedError()
 
+    def create_domain_parameters(self, key_type, attrs, local=False):
+        """
+        Create domain parameters.
+
+        The `local` parameter creates an object that is not created on the
+        HSM. This is useful if you only need the domain parameters to create
+        another object, and do not need a real object in the session.
+
+        :param KeyType key_type: Key type these parameters are for
+        :param dict(Attribute,*): Domain parameters (specific tp `key_type`)
+        :param local: if True, do not write these to the HSM.
+        :rtype: DomainParameters
+        """
+
+        raise NotImplementedError()
+
     def generate_key(self, key_type, key_length,
                      id=None, label=None,
                      store=True, capabilities=None,
-                     mechanism=None, mechanism_param=b'',
+                     mechanism=None, mechanism_param=None,
                      template=None):
         """
         Generate a single key (e.g. AES, DES).
@@ -315,11 +389,18 @@ class Session:
 
         An appropriate `mechanism` will be chosen for `key_type`
         (see :attr:`DEFAULT_GENERATE_MECHANISMS`) or this can be overridden.
-        Similarly the `capabilities` (see :attr:`DEFAULT_KEY_CAPABILITIES`).
+        Similarly for the `capabilities` (see
+        :attr:`DEFAULT_KEY_CAPABILITIES`).
 
         The `template` will extend the default template used to make the
         key.
 
+        Possible mechanisms and template attributes are defined by `PKCS #11
+        <http://docs.oasis-open.org/pkcs11/pkcs11-curr/v2.40/pkcs11-curr-v2.40.html>`_.
+        Invalid mechanisms or attributes should raise
+        :exc:`pkcs11.exceptions.MechanismInvalid` and
+        :exc:`pkcs11.exceptions.AttributeTypeInvalid` respectively.
+
         :param KeyType key_type: Key type (e.g. KeyType.AES)
         :param int key_length: Key length in bits (e.g. 256).
         :param bytes id: Key identifier.
@@ -330,26 +411,19 @@ class Session:
         :param bytes mechanism_param: Optional vector to the mechanism.
         :param dict(Attribute,*) template: Additional attributes.
 
-        :rtype: Key
+        :rtype: SecretKey
         """
         raise NotImplementedError()
 
     def generate_keypair(self, key_type, key_length,
                          id=None, label=None,
                          store=True, capabilities=None,
-                         mechanism=None, mechanism_param=b'',
+                         mechanism=None, mechanism_param=None,
                          public_template=None, private_template=None):
         """
         Generate a asymmetric keypair (e.g. RSA).
 
-        Keys should set at least `id` or `label`.
-
-        An appropriate `mechanism` will be chosen for `key_type`
-        (see :attr:`DEFAULT_GENERATE_MECHANISMS`) or this can be overridden.
-        Similarly the `capabilities` (see :attr:`DEFAULT_KEY_CAPABILITIES`).
-
-        The `template` will extend the default template used to make the
-        key.
+        See :meth:`generate_key` for more information.
 
         :param KeyType key_type: Key type (e.g. KeyType.AES)
         :param int key_length: Key length in bits (e.g. 256).
@@ -375,13 +449,42 @@ class Session:
 
     def generate_random(self, nbits):
         """
-        Generate `length` bytes of random or pseudo-random data (if supported).
+        Generate `length` bits of random or pseudo-random data (if supported).
 
         :param int nbits: Number of bits to generate.
         :rtype: bytes
         """
         raise NotImplementedError()
 
+    def digest(self, data, **kwargs):
+        """
+        Digest `data` using `mechanism`.
+
+        `data` can be a single value or an iterator.
+
+        :class:`Key` objects can also be digested, optionally interspersed
+        with :class:`bytes`.
+
+        :param data: Data to digest
+        :type data: str, bytes, Key or iter(bytes, Key)
+        :param Mechanism mechanism: digest mechanism
+        :param bytes mechanism_param: optional mechanism parameter
+
+        :rtype: bytes
+        """
+
+        # If data is a string, encode it now as UTF-8.
+        if isinstance(data, str):
+            data = data.encode('utf-8')
+
+        if isinstance(data, bytes):
+            return self._digest(data, **kwargs)
+
+        elif isinstance(data, Key):
+            data = (data,)
+
+        return self._digest_generator(data, **kwargs)
+
 
 class Object:
     """
@@ -389,6 +492,10 @@ class Object:
 
     Objects implement :meth:`__getitem__` and :meth:`__setitem__` to
     retrieve :class:`pkcs11.constants.Attribute` values on the object.
+    Valid attributes for an object are given in `PKCS #11
+    <http://docs.oasis-open.org/pkcs11/pkcs11-curr/v2.40/pkcs11-curr-v2.40.html>`_.
+    Invalid attributes should raise
+    :exc:`pkcs11.exceptions.AttributeTypeInvalid`.
     """
 
     object_class = None
@@ -441,6 +548,64 @@ class Object:
         raise NotImplementedError()
 
 
+class DomainParameters(Object):
+    """
+    PKCS#11 Domain Parameters.
+
+    Used to store domain parameters as part of the key generation step, e.g.
+    in DSA and Diffie-Hellman.
+    """
+
+    def __init__(self, session, handle, params=None):
+        super().__init__(session, handle)
+        self.params = params
+
+    def __getitem__(self, key):
+        if self._handle is None:
+            try:
+                return self.params[key]
+            except KeyError:
+                raise AttributeTypeInvalid
+        else:
+            return super().__getitem__(key)
+
+    def __setitem__(self, key, value):
+        if self._handle is None:
+            self.params[key] = value
+        else:
+            super().__setitem__(key, value)
+
+    @property
+    def key_type(self):
+        """
+        Key type (:class:`pkcs11.mechanisms.KeyType`) these parameters
+        can be used to generate.
+        """
+        return self[Attribute.KEY_TYPE]
+
+    def generate_keypair(self, id=None, label=None,
+                         store=False, capabilities=None,
+                         mechanism=None, mechanism_param=None,
+                         public_template=None, private_template=None):
+        """
+        Generate a key pair from these domain parameters (e.g. for
+        Diffie-Hellman.
+
+        See :meth:`Session.generate_key` for more information.
+
+        :param bytes id: Key identifier.
+        :param str label: Key label.
+        :param store: Store key on token (requires R/W session).
+        :param MechanismFlag capabilities: Key capabilities (or default).
+        :param Mechanism mechanism: Generation mechanism (or default).
+        :param bytes mechanism_param: Optional vector to the mechanism.
+        :param dict(Attribute,*) template: Additional attributes.
+
+        :rtype: (PublicKey, PrivateKey)
+        """
+        raise NotImplementedError()
+
+
 class Key(Object):
     """Base class for all key :class:`Object` types."""
 
@@ -490,6 +655,11 @@ class PublicKey(Key):
 
     object_class = ObjectClass.PUBLIC_KEY
 
+    @property
+    def key_length(self):
+        """Key length in bits."""
+        return self[Attribute.MODULUS_BITS]
+
 
 class PrivateKey(Key):
     """
@@ -499,6 +669,11 @@ class PrivateKey(Key):
 
     object_class = ObjectClass.PRIVATE_KEY
 
+    @property
+    def key_length(self):
+        """Key length in bits."""
+        return len(self[Attribute.MODULUS]) * 8
+
 
 class EncryptMixin(Object):
     """
@@ -519,11 +694,11 @@ class EncryptMixin(Object):
         is passed and the mechanism chosen does not support handling data
         in chunks, an exception will be raised.
 
-        Some mechanisms (including the default CBC mechanisms) require an
-        initialisation vector (of key length) to set the initial state of
-        the mechanism.  Pass this as `mechanism_param`. The initialisation
-        vector should contain quality random. This method will not return
-        the value of the initialisation parameter.
+        Some mechanisms (including the default CBC mechanisms) require
+        additional parameters, e.g. an initialisation vector [#]_, to
+        the mechanism.  Pass this as `mechanism_param`.
+        Documentation of these parameters is given specified in `PKCS #11
+        <http://docs.oasis-open.org/pkcs11/pkcs11-curr/v2.40/pkcs11-curr-v2.40.html>`_.
 
         When passing an iterable for data
         `buffer_size` must be sufficient to store the working buffer. An
@@ -545,15 +720,17 @@ class EncryptMixin(Object):
 
         ::
 
-            def encrypt_file(file_in, file_out):
+            def encrypt_file(file_in, file_out, buffer_size=8192):
 
                 with \\
                         open(file_in, 'rb') as input_, \\
                         open(file_out, 'wb') as output:
 
-                    chunks = iter(lambda: input_.read(8192), '')
+                    chunks = iter(lambda: input_.read(buffer_size), '')
 
-                    for chunk in key.encrypt(chunks, mechanism_param=iv):
+                    for chunk in key.encrypt(chunks,
+                                             mechanism_param=iv,
+                                             buffer_size=buffer_size):
                         output.write(chunk)
 
         :param data: data to encrypt
@@ -565,6 +742,11 @@ class EncryptMixin(Object):
         :param int buffer_size: size of the working buffer (for generators)
 
         :rtype: bytes or iter(bytes)
+
+        .. [#] The initialisation vector should contain quality random,
+            e.g. from :meth:`Session.generate_random`.
+            This method will not return the value of the initialisation
+            vector as part of the encryption. You must store that yourself.
         """
 
         # If data is a string, encode it now as UTF-8.
@@ -680,8 +862,149 @@ class VerifyMixin(Object):
 
 
 class WrapMixin(Object):
-    pass
+    """
+    This :class:`Object` supports the wrap capability.
+    """
+
+    def wrap_key(self, key,
+                 mechanism=None, mechanism_param=None):
+        """
+        Use this key to wrap (i.e. encrypt) `key` for export. Returns
+        an encrypted version of `key`.
+
+        `key` must have :attr:`Attribute.EXTRACTABLE` = True.
+
+        :param Key key: key to export
+        :param Mechanism mechanism: wrapping mechanism (or None for default).
+        :param bytes mechanism_param: mechanism parameter (if required)
+
+        :rtype: bytes
+        """
+        raise NotImplementedError()
 
 
 class UnwrapMixin(Object):
-    pass
+    """
+    This :class:`Object` supports the unwrap capability.
+    """
+
+    def unwrap_key(self, object_class, key_type, key_data,
+                   id=None, label=None,
+                   mechanism=None, mechanism_param=None,
+                   store=False, capabilities=None,
+                   template=None):
+        """
+        Use this key to unwrap (i.e. decrypt) and import `key_data`.
+
+        See :class:`Session.generate_key` for more information.
+
+        :param ObjectClass object_class: Object class to import as
+        :param KeyType key_type: Key type (e.g. KeyType.AES)
+        :param bytes key_data: Encrypted key to unwrap
+        :param bytes id: Key identifier.
+        :param str label: Key label.
+        :param store: Store key on token (requires R/W session).
+        :param MechanismFlag capabilities: Key capabilities (or default).
+        :param Mechanism mechanism: Generation mechanism (or default).
+        :param bytes mechanism_param: Optional vector to the mechanism.
+        :param dict(Attribute,*) template: Additional attributes.
+
+        :rtype: Key
+        """
+        raise NotImplementedError()
+
+
+class DeriveMixin(Object):
+    """
+    This :class:`Object` supports the derive capability.
+    """
+
+    def derive_key(self, key_type, key_length,
+                   id=None, label=None,
+                   store=False, capabilities=None,
+                   mechanism=None, mechanism_param=None,
+                   template=None):
+        """
+        Derive a new key from this key. Used to create session
+        keys from a PKCS key exchange.
+
+        Typically the mechanism, e.g. Diffie-Hellman, requires you
+        to specify the other party's piece of shared information as
+        the `mechanism_param`.  Some mechanisms require a tuple of data (see
+        :class:`pkcs11.mechanisms.Mechanism`).
+
+        See :class:`Session.generate_key` for more documentation on key
+        generation.
+
+        Diffie-Hellman example:
+
+        ::
+
+            # Diffie-Hellman domain parameters
+            # e.g. from RFC 3526, RFC 5114 or `openssl dhparam`
+            prime = [0xFF, ...]
+            base = [0x02]
+
+            parameters = session.create_domain_parameters(KeyType.DH, {
+                Attribute.PRIME: prime,
+                Attribute.BASE: base,
+            }, local=True)
+
+            # Alice generates a DH key pair from the public
+            # Diffie-Hellman parameters
+            public, private = parameters.generate_keypair()
+            alices_value = public[Attribute.VALUE]
+
+            # Bob generates a DH key pair from the same parameters.
+
+            # Alice exchanges public values with Bob...
+            # She sends `alices_value` and receives `bobs_value`.
+            # (Assuming Alice is doing AES CBC, she also needs to send an IV)
+
+            # Alice generates a session key with Bob's public value
+            # Bob will generate the same session key using Alice's value.
+            session_key = private.derive_key(
+                KeyType.AES, 128,
+                mechanism_param=bobs_value)
+
+        Elliptic-Curve Diffie-Hellman example:
+
+        ::
+
+            # DER encoded EC params, e.g. from OpenSSL
+            # openssl ecparam -outform der -name prime192v1 | base64
+            #
+            # Check what EC parameters the module supports with
+            # slot.get_module_info()
+            parameters = session.create_domain_parameters(KeyType.EC, {
+                Attribute.EC_PARAMS: b'...',
+            }, local=True)
+
+            # Alice generates a EC key pair, and gets her public value
+            public, private = parameters.generate_keypair()
+            alices_value = public[Attribute.EC_POINT]
+
+            # Bob generates a DH key pair from the same parameters.
+
+            # Alice exchanges public values with Bob...
+            # She sends `alices_value` and receives `bobs_value`.
+
+            # Alice generates a session key with Bob's public value
+            # Bob will generate the same session key using Alice's value.
+            session_key = private.derive_key(
+                KeyType.AES, 128,
+                mechanism_param=(KDF.NULL, None, bobs_value))
+
+        :param KeyType key_type: Key type (e.g. KeyType.AES)
+        :param int key_length: Key length in bits (e.g. 256).
+        :param bytes id: Key identifier.
+        :param str label: Key label.
+        :param store: Store key on token (requires R/W session).
+        :param MechanismFlag capabilities: Key capabilities (or default).
+        :param Mechanism mechanism: Generation mechanism (or default).
+        :param bytes mechanism_param: Optional vector to the mechanism.
+        :param dict(Attribute,*) template: Additional attributes.
+
+        :rtype: SecretKey
+        """
+        raise NotImplementedError()
